/***********************************************************************************************\
 * A.Keibels Geiger Counter 2019.
 * a.keibel@systragon.de
 * This sketch collects and displays the ticks of a connected Geiger-Counter 
 * on a connectd tiny 0.96'' Displays based on SSD1360. Use the adafruit library from the library manager. 
 * 
 *  
 * Pins:  * GND = GND,  * VCC = 5V,  * SCL = A5,  * SDA = A4, GEIGER_PIN = 2
 * The output on the screen shows the TICS PER MINUTE if the devider TPM2UNIT remains 1.0
 * a) The average result over the entire runtime.  
 * b) The average of only the last minute.
 * c) The average of the last 15 minutes.
 * d) The average of the last hour.
 * e) The average of the last six hours.
 * The display outputs starts on the first measured tick, assuming, that there was a tick right on powerup.
 *  
 * I have done this to find out about non statistical changes of radiation in my environement at different places.
 * I have found out, that during the day, the dose is around 20% increased compared to the night. 
 * I have also found out, that the dose is about 20% higher in the rooftop-floor, related to the basement(3 floors lower in my house).
 * 
 * This is achieved by a simple ring buffer with adjustable elements
 * Versions: 
 * V1.0        Only one ring buffer
 * V1.1        Multiple ring buffers, playing with the display
 * V1.3 (this) display a bargraphs with the buffer contents on the screen alternating in 4 seconds with the number values.
 * V1.4 (planning) move to ESP32/ESP8266 to go online for very long term statisticsa and alerting and mobile device access.
 * 
\*************************************************************************************************/
//#define DEBUG   // Enable this to see serial port outputs, but don't see outputs on the screen anymore because of lack of memory(?).

//#include <Adafruit_GFX.h>  // Include core graphics library for the display
#include <Adafruit_SSD1306.h>  // Include Adafruit_SSD1306 library to drive the display
#include <timer.h>

auto timer = timer_create_default(); // create a timer with default settings
Timer<> default_timer; // save as above

#define GEIGER_PIN    2   // Connect the TTL Geiger Counters output to this pin.

#define TPM2UNIT      1.0 // Dividing the tics per minute with this value converts the value to your desired unit.
                          // I am showing just the ticks, as I haven't found out about the type of geiger tupe yet.
// This struct collects tics for further processing
struct {
  long events;  // Number of events counted 
  long eps;     // Events per last second
  long eics;    // Events in current second
  long seconds; // Time of the measurement so far.
  long milliss; // Milliseconds at triggertime
  long rate;
}geiger;

#define SZEBUF 12   // How many entries shall the buffer have

typedef struct ringbuf{
//  unsigned int   elem[SZEBUF]; // Elements of the buffer
  unsigned int   *elem; // Elements of the buffer
  unsigned char  sze;   // Number of elements in the buffer
  char *         txt;   // NameText of display
  unsigned int   collector; // Collects the tics for the next element
  unsigned int   tb;        // Duration time of one element to collect tics (unit is seconds)
  unsigned int   sum;       // Latest collected sum of all tics inside the buffer
  double         result;    // Latest calculated overall tics/minute related to the overall buffer time.
};

//  The Ring-Buffers for the different times.
#define NUMBUFS 4
ringbuf TicBuf[NUMBUFS];

  //ringbuf tenMinutes;
  //ringbuf sixHours;
  //ringbuf oneDay;

void init_tbuffer(unsigned int period, unsigned char numElem, ringbuf * pbuf, char * txt){
  int i;
  pbuf->sze = numElem;
  pbuf->txt = txt;
  pbuf->elem = (unsigned int *)malloc( numElem* sizeof(unsigned int));
  pbuf->tb = period/pbuf->sze; //SZEBUF;
  pbuf->collector = 0;
  for(i=0; i<pbuf->sze; i++)pbuf->elem[i]=0;
}

bool toggle_led() {
  // This toggles the arduino on board LED for no reason ;-)  
  digitalWrite(LED_BUILTIN, !digitalRead(LED_BUILTIN)); // toggle the LED
  return true; // repeat? true
}

void ISR_ParticleCounter(){
  // This is the hardware interrupt service routine, triggerd by the Geiger Counter.
  geiger.events++; // Number of ovrall ticks
  geiger.eics++;   // Numbers of ticks in the currend second (collecting. reset in the seconds.timer)
  geiger.milliss = millis(); 
  geiger.rate= (long) 60000*geiger.events/geiger.milliss; // do I still need it?
  toggle_led(); // just for fun toggle the arduiono led on each event.
}

bool SecondsTick(void *p){
  // This is an elementary function, as in each second, the tics are collected in geiger.eics and stored in eps in this step.
  // All counters and ring-buffers collect the data from the events in the past second (geiger.eps) each second
  geiger.seconds++;  // Total runtime
  geiger.eps  = geiger.eics; // Save the collected tics of the past second...
  geiger.eics = 0;           // ... and reset it 
  
  // Let all buffers calculate on this base:
  int i;
  for(i=0; i < NUMBUFS; i++){
    update_tbuffer_per_second( &TicBuf[i]);
  }
  //  update_tbuffer_per_second( &oneMinute );
  //  update_tbuffer_per_second( &tenMinutes);
  //  update_tbuffer_per_second( &sixHours  );
  //  update_tbuffer_per_second( &oneDay    );
  return true;
}

void update_tbuffer_per_second(ringbuf * pbuf){
  // This function collects the tics in the actual part of the ring buffer 
  // and switches to the next element and organiztes the ringbuffer 
  // if the time of the element is reached then the next element is startet.
  // The duration which is covered by a buffer entry is pbuf->tb
  // Input parameter are the overall seconds geiger.seconds 
  // First Modulo with the time of the complete buffer, the clock then starts over after the complete buffer is completely filled.
  // Then DIV this again with the duration of each element to get the the actual index.
  int index, index_new, i;
  // Calculate the index of the current buffer-element-index related to the last second of the overall runtime. 
  index    = ((geiger.seconds  ) % (pbuf->tb * pbuf->sze) ) / pbuf->tb;
  // If the next index is different, then do some organisational stuff. see below.
  index_new= ((geiger.seconds+1) % (pbuf->tb * pbuf->sze) ) / pbuf->tb;

  // Ad the tics, collected in the last second to the collector for this counter.
  pbuf->collector  += geiger.eps;

  if(geiger.seconds < pbuf->sze * pbuf->tb){
    // The buffer has not been filled yet. just collect the collectors tics: 
    pbuf->elem[index] = pbuf->collector; 
    if(index_new != index){
      // If the bufferindex is going to change,-> reset the counter, to collect for the next index, then.
      pbuf->collector=0;
    }
  }else{
    // The buffer is full  
    if(index_new != index){
      // The index will change on the nexst second,
      // Take over all collected ticks into the buffer element:
      pbuf->elem[index] = pbuf->collector; 
      pbuf->collector   = 0; //  and reset the collector to start over for the next element
    }else{
      // the collector collects the tics in the seconds events. so nothing to do here.
    }
  }
  // How many tics are in the entire buffer?
  pbuf->sum = 0; 
  for(i = 0; i<pbuf->sze; i++){  pbuf->sum += pbuf->elem[i];
  #ifdef DEBUG        
      Serial.print(pbuf->elem[i]);Serial.print(", ");
  #endif    
  }

  // Calculate the tics over the entire buffertime in tics per second: 
  if(geiger.seconds < pbuf->sze * pbuf->tb){
    // If the buffer is not full yet, then use the entire runtime, and not the buffer time:
    pbuf->result = (double)60.0* pbuf->sum / geiger.seconds;
    #ifdef DEBUG        
        Serial.print(pbuf->result);Serial.print(" *");
    #endif    
  }else{
    // Buffer is full, so use the entire buffer time:
    pbuf->result = (double)60.0* pbuf->sum / (pbuf->sze * pbuf->tb);
    #ifdef DEBUG        
        Serial.print(pbuf->result);Serial.print("");    
    #endif    
  }
#ifdef DEBUG      
  Serial.println(" ");
#endif  
  
}

Adafruit_SSD1306 display(128, 64);  // Create display

//#include <Fonts/FreeMonoBold12pt7b.h>  // Add a custom font
#include <Fonts/FreeMono9pt7b.h>       // Add a custom font

void setup(){  // Start of setup
  #ifdef DEBUG   
    Serial.begin(115200);
  #endif  
  // Init the data stuff for the Geiger Counter and Ring-Bufffers for different time- delays
  pinMode(LED_BUILTIN, OUTPUT); // set LED pin to OUTPUT

  
  delay(100);  // This delay is needed to let the display to initialize
  // call the SecondsTick function every 1000 millis (1 second)
  timer.every(1000, SecondsTick, (void *)10);
  
  // Initialize display with the I2C address of 0x3C
  if(!display.begin(SSD1306_SWITCHCAPVCC, 0x3C)){
  #ifdef DEBUG     
      Serial.println(F("SSD1306 allocation failed"));
      // For some reason this fails exactly in this case, if the Serial.print is used. 
  #endif    
     while(1){
         // Blink the LED to show, that something is wrong. 
         digitalWrite(LED_BUILTIN, !digitalRead(LED_BUILTIN)); delay(250);
     }
  }
  
  display.clearDisplay();  // Clear the buffer

  display.setTextColor(WHITE);  // Set color of the text

  display.setRotation(0);  // Set orientation. Goes from 0, 1, 2 or 3

  display.setTextWrap(false);  // By default, long lines of text are set to automatically “wrap” back to the leftmost column.
                               // To override this behavior (so text will run off the right side of the display - useful for
                               // scrolling marquee effects), use setTextWrap(false). The normal wrapping behavior is restored
                               // with setTextWrap(true).
  display.dim(0);  //Set brightness (0 is maximun and 1 is a little dim)


  geiger.events  = 0;
  geiger.seconds = 0;
  geiger.milliss = 1;
  geiger.rate    = 1;
  pinMode(GEIGER_PIN,INPUT);
  attachInterrupt(digitalPinToInterrupt(GEIGER_PIN), ISR_ParticleCounter, FALLING);
  
  init_tbuffer(60      , 10, &TicBuf[0], "1MI");
  init_tbuffer(60*10   ,  6, &TicBuf[1], "10M");
  init_tbuffer(60*60*6 ,  6, &TicBuf[2], "6Hr");
  init_tbuffer(60*60*24, 24, &TicBuf[3], "1Dy");
}  // End of setup

void drawChart(int x, int y, int h, ringbuf * pbuf){
  /*  This function draws a normalized chart of one buffer on the screen.
   *  In a way, that only the variation is drawn. The minimum value of the smalles entry 
   *  in the ringbuffer is substracted before drawing. So we see only the changes over the time of the buffer
   *  And the range is scaled, so that it fits into the given height h of the chart 
   *  With this we see clearly if the radiation has decreased or increased. 
   *  
  */
  int mi,ma,i,w,v,idx,id;
  w =(128-x)/pbuf->sze; // Width of one bar
  ma=pbuf->elem[0]; // Start values for maximum minimum search for scaling
  mi=ma;
  for(i=0; i < pbuf->sze;i++){
    if (pbuf->elem[i]>ma) ma= pbuf->elem[i];
    if (pbuf->elem[i]<mi) mi= pbuf->elem[i];
  }
  // Get the lastest index, to put it first
  idx    = ((geiger.seconds  ) % (pbuf->tb * pbuf->sze) ) / pbuf->tb + pbuf->sze; 
  for(i=0; i < pbuf->sze;i++){     
      
      id = idx % pbuf->sze; // increment idx and modulo it with the size of the buffer
      v  = h*(pbuf->elem[id]-mi)/(ma-mi);
      idx--;
      display.fillRect(x + i*w, y+h-v, w, v+1, WHITE);  
  }  
}
void loop(){  // Start of loop
  int i;
  timer.tick(); // tick the timer
  char string[10];  // Create a character array of some characters
  display.clearDisplay();  // Clear the display so we can refresh

  display.setFont(&FreeMono9pt7b);  // Set a custom font
  display.setTextSize(0);  // Set text size. We are using a custom font so you should always use the text size of 0
  display.setCursor(0, 10);  // (x,y)
  display.println("ALL");  // Text or value to print

  for (i=0; i< NUMBUFS; i++){
      // Print the Names on the screen
      display.setCursor(0, 22+i*12);  // (x,y)
      display.println(TicBuf[i].txt);  // Text or value to print 
  }
  if (geiger.seconds/2 % 2 == 0){
    // Draw the charts onto the screen
    for (i=0; i< NUMBUFS; i++){
      drawChart(35, 12+i*12, 10, &TicBuf[i]);
    }
  }else{
    for (i=0; i< NUMBUFS; i++){
        // Print the values on the screen
        dtostrf(TicBuf[i].result/TPM2UNIT, 9, 2, string);  
        display.setCursor(30, 22+i*12); display.println(string); 
      }
  }
  
  // Draw rectangle:
  //display.drawRect(40, 1, 87, 25, WHITE);  // Draw rectangle (x,y,width,height,color)
                                             // It draws from the location to down-right
  dtostrf((double)(60000*geiger.events/millis())/TPM2UNIT, 9, 2, string);  // (<variable>,<amount of digits we are going to use>,<amount of decimal digits>,<string name>)
  // Display the counter rate per minute
  display.setCursor(30, 10); display.println(string);  // Text or value to print
  display.display();  // Print everything we set previously

}  // End of loop